@@ -27,6 +27,7 @@
 #include <fcntl.h>
 #include <syslog.h>
 #include <stdlib.h>
+#include <sys/mount.h>
 
 #ifndef __GNUC__
 #error I need GCC to work
@@ -36,20 +37,29 @@
 #define MY_NAME "nbd_client"
 #include "cliserv.h"
 
-int opennet(char *name, int port) {
+int opennet(char *name, int port, int sdp) {
 	int sock;
 	struct sockaddr_in xaddrin;
 	int xaddrinlen = sizeof(xaddrin);
 	struct hostent *hostn;
+	int af;
 
 	hostn = gethostbyname(name);
 	if (!hostn)
 		err("Gethostname failed: %h\n");
 
-	if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
+	af = AF_INET;
+	if(sdp) {
+#ifdef WITH_SDP
+		af = AF_INET_SDP;
+#else
+		err("Can't do SDP: I was not compiled with SDP support!");
+#endif
+	}
+	if ((sock = socket(af, SOCK_STREAM, IPPROTO_TCP)) < 0)
 		err("Socket failed: %m");
 
-	xaddrin.sin_family = AF_INET;
+	xaddrin.sin_family = af;
 	xaddrin.sin_port = htons(port);
 	xaddrin.sin_addr.s_addr = *((int *) hostn->h_addr);
 	if ((connect(sock, (struct sockaddr *) &xaddrin, xaddrinlen) < 0))
@@ -59,7 +69,7 @@ int opennet(char *name, int port) {
 	return sock;
 }
 
-u64 negotiate(int sock, int blocksize) {
+void negotiate(int sock, u64 *rsize64, u32 *flags) {
 	u64 magic, size64;
 	char buf[256] = "\0\0\0\0\0\0\0\0\0";
 
@@ -96,15 +106,20 @@ u64 negotiate(int sock, int blocksize) {
 		printf("size = %lu", (unsigned long)(size64));
 #endif
 
-	if (read(sock, &buf, 128) < 0)
+	if (read(sock, flags, sizeof(*flags)) < 0)
 		err("Failed/4: %m\n");
+	*flags = ntohl(*flags);
+
+	if (read(sock, &buf, 124) < 0)
+		err("Failed/5: %m\n");
 	printf("\n");
 
-	return size64;
+	*rsize64 = size64;
 }
 
-void setsizes(int nbd, u64 size64, int blocksize) {
+void setsizes(int nbd, u64 size64, int blocksize, u32 flags) {
 	unsigned long size;
+	int read_only = (flags & NBD_FLAG_READ_ONLY) ? 1 : 0;
 
 #ifdef NBD_SET_SIZE_BLOCKS
 	if (size64/blocksize > (~0UL >> 1))
@@ -129,6 +144,19 @@ void setsizes(int nbd, u64 size64, int blocksize) {
 #endif
 
 	ioctl(nbd, NBD_CLEAR_SOCK);
+
+	if (ioctl(nbd, BLKROSET, (unsigned long) &read_only) < 0)
+		err("Unable to set read-only attribute for device");
+}
+
+void set_timeout(int nbd, int timeout) {
+#ifdef NBD_SET_TIMEOUT
+	if (timeout) {
+		if (ioctl(nbd, NBD_SET_TIMEOUT, (unsigned long)timeout) < 0)
+			err("Ioctl NBD_SET_TIMEOUT failed: %m\n");
+		fprintf(stderr, "timeout=%d\n", timeout);
+	}
+#endif
 }
 
 void finish_sock(int sock, int nbd, int swap) {
@@ -151,14 +179,17 @@ int main(int argc, char *argv[]) {
 	char *hostname, *nbddev;
 	int swap=0;
 	int cont=0;
+	int timeout=0;
+	int sdp=0;
 	u64 size64;
+	u32 flags;
 
 	logging();
 
 	if (argc < 3) {
 	errmsg:
 		fprintf(stderr, "nbd-client version %s\n", PACKAGE_VERSION);
-		fprintf(stderr, "Usage: nbd-client [bs=blocksize] host port nbd_device [-swap] [-persist]\n");
+		fprintf(stderr, "Usage: nbd-client [bs=blocksize] [timeout=sec] host port nbd_device [-swap] [-persist]\n");
 		fprintf(stderr, "Or   : nbd-client -d nbd_device\n");
 		fprintf(stderr, "Default value for blocksize is 1024 (recommended for ethernet)\n");
 		fprintf(stderr, "Allowed values for blocksize are 512,1024,2048,4096\n"); /* will be checked in kernel :) */
@@ -196,6 +227,11 @@ int main(int argc, char *argv[]) {
 		++argv; --argc; /* skip blocksize */
 	}
 
+	if (strncmp(argv[0], "timeout=", 8)==0) {
+		timeout=atoi(argv[0]+8);
+		++argv; --argc; /* skip timeout */
+	}
+	
 	if (argc==0) goto errmsg;
 	hostname=argv[0];
 	++argv; --argc; /* skip hostname */
@@ -205,30 +241,38 @@ int main(int argc, char *argv[]) {
 	++argv; --argc; /* skip port */
 
 	if (argc==0) goto errmsg;
-	sock = opennet(hostname, port);
 	nbddev = argv[0];
 	nbd = open(nbddev, O_RDWR);
 	if (nbd < 0)
 	  err("Can not open NBD: %m");
 	++argv; --argc; /* skip device */
 
-	if (argc>2) goto errmsg;
-	if (argc!=0) {
+	if (argc>3) goto errmsg;
+	if (argc) {
 		if(strncmp(argv[0], "-swap", 5)==0) {
 			swap=1;
 			++argv;--argc;
 		}
 	}
-	if (argc!=0) {
+	if (argc) {
 		if(strncmp(argv[0], "-persist", 8)==0) {
 			cont=1;
 			++argv;--argc;
 		}
 	}
+	if (argc) {
+		if(strncmp(argv[0], "-sdp", 4)==0) {
+			sdp=1;
+			++argv;--argc;
+		}
+	}
+	if(argc) goto errmsg;
+	sock = opennet(hostname, port, sdp);
 	argv=NULL; argc=0; /* don't use it later suddenly */
 
-	size64 = negotiate(sock, blocksize);
-	setsizes(nbd, size64, blocksize);
+	negotiate(sock, &size64, &flags);
+	setsizes(nbd, size64, blocksize, flags);
+	set_timeout(nbd, timeout);
 	finish_sock(sock, nbd, swap);
 
 	/* Go daemon */
@@ -248,14 +292,21 @@ int main(int argc, char *argv[]) {
 				cont=0;
 			} else {
 				if(cont) {
+					u64 new_size;
+					u32 new_flags;
+
 					fprintf(stderr, " Reconnecting\n");
 					close(sock); close(nbd);
-					sock = opennet(hostname, port);
+					sock = opennet(hostname, port, sdp);
 					nbd = open(nbddev, O_RDWR);
-					if(size64!=negotiate(sock,blocksize)) {
+					negotiate(sock, &new_size, &new_flags);
+					if (size64 != new_size) {
 						err("Size of the device changed. Bye");
 					}
-					setsizes(nbd, size64, blocksize);
+					setsizes(nbd, size64, blocksize,
+								new_flags);
+
+					set_timeout(nbd, timeout);
 					finish_sock(sock,nbd,swap);
 				}
 			}
